@@ -1,15 +1,33 @@
 import atexit
+import getpass
+import json
 import logging
 import os
 import sys
-import json
-import getpass
-from mintapi import constants
-import keyring
-import configargparse
 
+import configargparse
+import keyring
 from mintapi.api import Mint
+from mintapi.constants import (
+    ACCOUNT_KEY,
+    BILL_KEY,
+    BUDGET_KEY,
+    CATEGORY_KEY,
+    CREDIT_REPORT_KEY,
+    CREDIT_SCORE_KEY,
+    CSV_FORMAT,
+    INVESTMENT_KEY,
+    JSON_FORMAT,
+    MFA_VIA_EMAIL,
+    MFA_VIA_SMS,
+    MFA_VIA_SOFT_TOKEN,
+    NET_WORTH_KEY,
+    TRANSACTION_KEY,
+    TRENDS_KEY,
+)
+from mintapi.filters import DateFilter
 from mintapi.signIn import get_email_code
+from mintapi.trends import ReportView
 from pandas import json_normalize
 
 logger = logging.getLogger("mintapi")
@@ -174,8 +192,8 @@ def parse_arguments(args):
         (
             ("--format",),
             {
-                "choices": [constants.JSON_FORMAT, constants.CSV_FORMAT],
-                "default": constants.JSON_FORMAT,
+                "choices": [JSON_FORMAT, CSV_FORMAT],
+                "default": JSON_FORMAT,
                 "help": "The format used to return data.",
             },
         ),
@@ -194,6 +212,13 @@ def parse_arguments(args):
             ("--imap-test",),
             {"action": "store_true", "help": "Test imap login and retrieval."},
         ),
+        (
+            ("--intuit-account",),
+            {
+                "default": None,
+                "help": "Specify an override of the default intuit account for accessing Mint",
+            },
+        ),
         (
             ("--investments",),
             {
@@ -229,9 +254,9 @@ def parse_arguments(args):
             ("--mfa-method",),
             {
                 "choices": [
-                    constants.MFA_VIA_SMS,
-                    constants.MFA_VIA_EMAIL,
-                    constants.MFA_VIA_SOFT_TOKEN,
+                    MFA_VIA_SMS,
+                    MFA_VIA_EMAIL,
+                    MFA_VIA_SOFT_TOKEN,
                 ],
                 "default": None,
                 "help": "The MFA method to automate.",
@@ -283,10 +308,46 @@ def parse_arguments(args):
                 "help": "Earliest date for transactions to be retrieved from. Used with --transactions. Format: mm/dd/yy",
             },
         ),
+        (
+            ("--transaction-date-filter",),
+            {
+                "type": int,
+                "default": DateFilter.Options.ALL_TIME,
+                "dest": "transaction_date_filter",
+                "help": "The date window for which to generate your transaction search.  Default is All Time.",
+            },
+        ),
         (
             ("--transactions", "-t"),
             {"action": "store_true", "default": False, "help": "Retrieve transactions"},
         ),
+        (
+            ("--trends",),
+            {
+                "action": "store_true",
+                "dest": "trends",
+                "default": False,
+                "help": "Retrieve trend data related to your financial information",
+            },
+        ),
+        (
+            ("--trend-report-type",),
+            {
+                "type": int,
+                "default": ReportView.Options.SPENDING_TIME,
+                "dest": "trend_report_type",
+                "help": "The type of report for which to generate trend analysis.  Default is Spending Over Time.",
+            },
+        ),
+        (
+            ("--trend-date-filter",),
+            {
+                "type": int,
+                "default": DateFilter.Options.THIS_MONTH,
+                "dest": "trend_date_filter",
+                "help": "The date window for which to generate your trend analysis.  Default is This Month.",
+            },
+        ),
         (
             ("--use-chromedriver-on-path",),
             {
@@ -294,6 +355,13 @@ def parse_arguments(args):
                 "help": "Whether to use the chromedriver on PATH, instead of downloading a local copy.",
             },
         ),
+        (
+            ("--use-rest-client",),
+            {
+                "action": "store_true",
+                "help": "Prefer the REST client for API calls. This will be the default in a future release",
+            },
+        ),
         (
             ("--wait_for_sync_timeout",),
             {
@@ -313,10 +381,10 @@ def parse_arguments(args):
     return cmdline.parse_args(args)
 
 
-def handle_password(type, prompt, email, password, use_keyring=False):
+def handle_password(name, prompt, email, password, use_keyring=False):
     if use_keyring and not password:
         # If we don't yet have a password, try prompting for it
-        password = keyring.get_password(type, email)
+        password = keyring.get_password(name, email)
 
     if not password:
         # If we still don't have a password, prompt for it
@@ -324,35 +392,35 @@ def handle_password(type, prompt, email, password, use_keyring=False):
 
     if use_keyring:
         # If keyring option is specified, save the password in the keyring
-        keyring.set_password(type, email, password)
+        keyring.set_password(name, email, password)
 
     return password
 
 
-def format_filename(options, type):
+def format_filename(options, name):
     if options.filename is None:
         filename = None
     else:
-        filename = "{}_{}.{}".format(options.filename, type.lower(), options.format)
+        filename = "{}_{}.{}".format(options.filename, name.lower(), options.format)
     return filename
 
 
-def output_data(options, data, type, attention_msg=None):
-    filename = format_filename(options, type)
+def output_data(options, data, name, attention_msg=None):
+    filename = format_filename(options, name)
     if filename is None:
-        if options.format == constants.CSV_FORMAT:
+        if options.format == CSV_FORMAT:
             print(json_normalize(data).to_csv(index=False))
         else:
             print(json.dumps(data, indent=2))
         # NOTE: While this logic is here, unless validate_file_extensions
         #       allows for other data types to export to CSV, this will
         #       only include investment data.
-    elif options.format == constants.CSV_FORMAT:
+    elif options.format == CSV_FORMAT:
         # NOTE: Currently, investment_data, which is a flat JSON, is the only
         #       type of data that uses this section.  So, if we open this up to
         #       other non-flat JSON data, we will need to revisit this.
         json_normalize(data).to_csv(filename, index=False)
-    elif options.format == constants.JSON_FORMAT:
+    elif options.format == JSON_FORMAT:
         with open(filename, "w+") as f:
             json.dump(data, f, indent=2)
 
@@ -375,6 +443,9 @@ def main():
     imap_account = options.imap_account
     imap_password = options.imap_password
     mfa_method = options.mfa_method
+    report_type = ReportView.Options(options.trend_report_type)
+    trend_date_filter = DateFilter.Options(options.trend_date_filter)
+    transaction_date_filter = DateFilter.Options(options.transaction_date_filter)
 
     if not email:
         # If the user did not provide an e-mail, prompt for it
@@ -399,6 +470,7 @@ def main():
             options.bills,
             options.budgets,
             options.transactions,
+            options.trends,
             options.net_worth,
             options.credit_score,
             options.credit_report,
@@ -415,8 +487,8 @@ def main():
         session_path = options.session_path
 
     mint = Mint(
-        email,
-        password,
+        email=email,
+        password=password,
         mfa_method=mfa_method,
         mfa_token=options.mfa_token,
         session_path=session_path,
@@ -425,10 +497,12 @@ def main():
         imap_password=imap_password,
         imap_server=options.imap_server,
         imap_folder=options.imap_folder,
+        intuit_account=options.intuit_account,
         wait_for_sync=not options.no_wait_for_sync,
         wait_for_sync_timeout=options.wait_for_sync_timeout,
         fail_if_stale=options.fail_if_stale,
         use_chromedriver_on_path=options.use_chromedriver_on_path,
+        use_rest_client=options.use_rest_client,
         chromedriver_download_path=options.chromedriver_download_path,
         beta=options.beta,
     )
@@ -449,52 +523,75 @@ def main():
     if options.attention:
         attention_msg = mint.get_attention()
 
+    if options.trends:
+        data = mint.get_trend_data(
+            report_type=report_type,
+            date_filter=trend_date_filter,
+            start_date=options.start_date,
+            end_date=options.end_date,
+            category_ids=None,
+            tag_ids=None,
+            descriptions=None,
+            account_ids=None,
+            match_all_filters=True,
+            limit=options.limit,
+            offset=0,
+        )
+        output_data(options, data, TRENDS_KEY, attention_msg)
+
     if options.accounts:
         data = mint.get_account_data(limit=options.limit)
-        output_data(options, data, constants.ACCOUNT_KEY, attention_msg)
+        output_data(options, data, ACCOUNT_KEY, attention_msg)
 
     if options.bills:
-        data = mint.get_bills()
-        output_data(options, data, constants.BILL_KEY, attention_msg)
+        data = mint.get_bills_data()
+        output_data(options, data, BILL_KEY, attention_msg)
 
     if options.budgets:
         data = mint.get_budget_data(limit=options.limit)
-        output_data(options, data, constants.BUDGET_KEY, attention_msg)
+        output_data(options, data, BUDGET_KEY, attention_msg)
     elif options.budget_hist:
         data = mint.get_budget_data(limit=options.limit, hist=12)
-        output_data(options, data, constants.BUDGET_KEY, attention_msg)
+        output_data(options, data, BUDGET_KEY, attention_msg)
 
     if options.transactions:
         data = mint.get_transaction_data(
-            limit=options.limit,
+            date_filter=transaction_date_filter,
             start_date=options.start_date,
             end_date=options.end_date,
+            category_ids=None,
+            tag_ids=None,
+            descriptions=None,
+            account_ids=None,
+            match_all_filters=True,
             include_investment=options.include_investment,
             remove_pending=options.show_pending,
+            limit=options.limit,
+            offset=0,
         )
-        output_data(options, data, constants.TRANSACTION_KEY, attention_msg)
+        output_data(options, data, TRANSACTION_KEY, attention_msg)
 
     if options.categories:
         data = mint.get_category_data(
             limit=options.limit,
         )
-        output_data(options, data, constants.CATEGORY_KEY, attention_msg)
+        output_data(options, data, CATEGORY_KEY, attention_msg)
 
     if options.investments:
         data = mint.get_investment_data(
             limit=options.limit,
         )
-        output_data(options, data, constants.INVESTMENT_KEY, attention_msg)
+        output_data(options, data, INVESTMENT_KEY, attention_msg)
 
     if options.net_worth:
         data = mint.get_net_worth_data()
         formatted_data = {"net_worth": data}
-        output_data(options, formatted_data, constants.NET_WORTH_KEY, attention_msg)
+        output_data(options, formatted_data, NET_WORTH_KEY, attention_msg)
 
     if options.credit_score:
         data = mint.get_credit_score_data()
         formatted_data = {"credit_score": data}
-        output_data(options, formatted_data, constants.CREDIT_SCORE_KEY, attention_msg)
+        output_data(options, formatted_data, CREDIT_SCORE_KEY, attention_msg)
 
     if options.credit_report:
         data = mint.get_credit_report_data(
@@ -503,4 +600,4 @@ def main():
             exclude_accounts=options.exclude_accounts,
             exclude_utilization=options.exclude_utilization,
         )
-        output_data(options, data, constants.CREDIT_REPORT_KEY, attention_msg)
+        output_data(options, data, CREDIT_REPORT_KEY, attention_msg)
