@@ -1,17 +1,13 @@
+import os
 from datetime import datetime
 from mintapi import constants, exceptions
 import email
 import email.header
 import imaplib
-import io
 import logging
-import os
 import re
-import requests
-import subprocess
 import sys
 import time
-import zipfile
 
 from selenium.common.exceptions import (
     ElementNotInteractableException,
@@ -19,9 +15,11 @@ from selenium.common.exceptions import (
     NoSuchElementException,
     StaleElementReferenceException,
     TimeoutException,
+    WebDriverException,
 )
 from selenium.webdriver import ChromeOptions
 from selenium.webdriver.common.by import By
+from selenium.webdriver.common.service import Service as ChromeService
 from selenium.webdriver.support import expected_conditions
 from selenium.webdriver.support.ui import WebDriverWait
 from seleniumrequests import Chrome
@@ -57,16 +55,16 @@ MFA_METHODS = [
     },
     {
         constants.MFA_METHOD_LABEL: constants.MFA_VIA_EMAIL,
-        SELECT_CSS_SELECTORS_LABEL: "#ius-label-mfa-email-otp, #ius-mfa-email-otp-card-challenge, #ius-sublabel-mfa-email-otp",
+        SELECT_CSS_SELECTORS_LABEL: '#ius-label-mfa-email-otp, #ius-mfa-email-otp-card-challenge, #ius-sublabel-mfa-email-otp, [data-testid="challengePickerOption_EMAIL_OTP"]',
         INPUT_CSS_SELECTORS_LABEL: "#ius-mfa-confirm-code",
-        SPAN_CSS_SELECTORS_LABEL: '[data-testid="VerifyOtpHeaderText"]',
+        SPAN_CSS_SELECTORS_LABEL: '[data-testid="VerifyOtpHeaderText"], #VerifyOtpHeader',
         BUTTON_CSS_SELECTORS_LABEL: '#ius-mfa-otp-submit-btn, [data-testid="VerifyOtpSubmitButton"]',
     },
     {
         constants.MFA_METHOD_LABEL: constants.MFA_VIA_SMS,
-        SELECT_CSS_SELECTORS_LABEL: "#ius-mfa-sms-otp-card-challenge",
+        SELECT_CSS_SELECTORS_LABEL: '#ius-mfa-sms-otp-card-challenge, [data-testid="challengePickerOption_SMS_OTP"]',
         INPUT_CSS_SELECTORS_LABEL: "#ius-mfa-confirm-code",
-        SPAN_CSS_SELECTORS_LABEL: '[data-testid="VerifyOtpHeaderText"]',
+        SPAN_CSS_SELECTORS_LABEL: '[data-testid="VerifyOtpHeaderText"], #VerifyOtpHeader',
         BUTTON_CSS_SELECTORS_LABEL: '#ius-mfa-otp-submit-btn, [data-testid="VerifyOtpSubmitButton"]',
     },
 ]
@@ -108,7 +106,7 @@ def get_email_code(imap_account, imap_password, imap_server, imap_folder, delete
             count = count + 1
             if count > 3:
                 break
-            rv, data = imap_client.fetch(num, "(RFC822)")
+            rv, data = imap_client.fetch(num, "(BODY.PEEK[])")
             if rv != "OK":
                 raise RuntimeError("Unable to complete due to error message: " + rv)
 
@@ -125,8 +123,13 @@ def get_email_code(imap_account, imap_password, imap_server, imap_folder, delete
             if not re.search("do_not_reply@intuit.com", frm, re.IGNORECASE):
                 continue
 
-            if not re.search("Your Mint Account", subject, re.IGNORECASE):
+            p = re.search(r"(\d\d\d\d\d\d) Mint code", subject)
+            if p:
+                code = p.group(1)
+            elif not re.search("Your Mint (code|Account)", subject, re.IGNORECASE):
                 continue
+            else:
+                code = ""
 
             date_tuple = email.utils.parsedate_tz(msg["Date"])
             if date_tuple:
@@ -143,13 +146,15 @@ def get_email_code(imap_account, imap_password, imap_server, imap_folder, delete
 
             logger.debug("DEBUG: EMAIL HEADER OK")
 
-            body = str(msg)
-
-            p = re.search(r"Verification code:<.*?(\d\d\d\d\d\d)$", body, re.S | re.M)
-            if p:
-                code = p.group(1)
-            else:
-                logger.error("FAIL1")
+            if code == "":
+                body = next(msg.walk()).get_payload(None, True).decode()
+                p = re.search(
+                    r"Verification code:<.*?(\d\d\d\d\d\d)\b", body, re.S | re.M
+                )
+                if p:
+                    code = p.group(1)
+                else:
+                    logger.error("FAIL1")
 
             logger.debug("DEBUG: CODE FROM EMAIL:", code)
 
@@ -173,112 +178,6 @@ def get_email_code(imap_account, imap_password, imap_server, imap_folder, delete
     return code
 
 
-CHROME_DRIVER_BASE_URL = "https://chromedriver.storage.googleapis.com/"
-CHROME_DRIVER_DOWNLOAD_PATH = "{version}/chromedriver_{arch}.zip"
-CHROME_DRIVER_LATEST_RELEASE = "LATEST_RELEASE"
-CHROME_ZIP_TYPES = {
-    "linux": "linux64",
-    "linux2": "linux64",
-    "darwin": "mac64",
-    "win32": "win32",
-    "win64": "win32",
-}
-version_pattern = re.compile(
-    "(?P<version>(?P<major>\\d+)\\.(?P<minor>\\d+)\\."
-    "(?P<build>\\d+)\\.(?P<patch>\\d+))"
-)
-
-
-def get_chrome_driver_url(version, arch):
-    return CHROME_DRIVER_BASE_URL + CHROME_DRIVER_DOWNLOAD_PATH.format(
-        version=version, arch=CHROME_ZIP_TYPES.get(arch)
-    )
-
-
-def get_chrome_driver_major_version_from_executable(local_executable_path):
-    # Note; --version works on windows as well.
-    # check_output fails if running from a thread without a console on win10.
-    # To protect against this use explicit pipes for STDIN/STDERR.
-    # See: https://github.com/pyinstaller/pyinstaller/issues/3392
-    with open(os.devnull, "wb") as devnull:
-        version = subprocess.check_output(
-            [local_executable_path, "--version"], stderr=devnull, stdin=devnull
-        )
-        version_match = version_pattern.search(version.decode())
-        if not version_match:
-            return None
-        return version_match.groupdict()["major"]
-
-
-def get_latest_chrome_driver_version():
-    """Returns the version of the latest stable chromedriver release."""
-    latest_url = CHROME_DRIVER_BASE_URL + CHROME_DRIVER_LATEST_RELEASE
-    latest_request = requests.get(latest_url)
-
-    if latest_request.status_code != 200:
-        raise RuntimeError(
-            "Error finding the latest chromedriver at {}, status = {}".format(
-                latest_url, latest_request.status_code
-            )
-        )
-    return latest_request.text
-
-
-def get_stable_chrome_driver(download_directory=os.getcwd()):
-    chromedriver_name = "chromedriver"
-    if sys.platform in ["win32", "win64"]:
-        chromedriver_name += ".exe"
-
-    local_executable_path = os.path.join(download_directory, chromedriver_name)
-
-    latest_chrome_driver_version = get_latest_chrome_driver_version()
-    version_match = version_pattern.match(latest_chrome_driver_version)
-    latest_major_version = None
-    if not version_match:
-        logger.error(
-            "Cannot parse latest chrome driver string: {}".format(
-                latest_chrome_driver_version
-            )
-        )
-    else:
-        latest_major_version = version_match.groupdict()["major"]
-    if os.path.exists(local_executable_path):
-        major_version = get_chrome_driver_major_version_from_executable(
-            local_executable_path
-        )
-        if major_version == latest_major_version or not latest_major_version:
-            # Use the existing chrome driver, as it's already the latest
-            # version or the latest version cannot be determined at the moment.
-            return local_executable_path
-        logger.info("Removing old version {} of Chromedriver".format(major_version))
-        os.remove(local_executable_path)
-
-    if not latest_chrome_driver_version:
-        logger.critical(
-            "No local chrome driver found and cannot parse the latest chrome "
-            "driver on the internet. Please double check your internet "
-            "connection, then ask for assistance on the github project."
-        )
-        return None
-    logger.info(
-        "Downloading version {} of Chromedriver".format(latest_chrome_driver_version)
-    )
-    zip_file_url = get_chrome_driver_url(latest_chrome_driver_version, sys.platform)
-    request = requests.get(zip_file_url)
-
-    if request.status_code != 200:
-        raise RuntimeError(
-            "Error finding chromedriver at {}, status = {}".format(
-                zip_file_url, request.status_code
-            )
-        )
-
-    zip_file = zipfile.ZipFile(io.BytesIO(request.content))
-    zip_file.extractall(path=download_directory)
-    os.chmod(local_executable_path, 0o755)
-    return local_executable_path
-
-
 def _create_web_driver_at_mint_com(
     headless=False,
     session_path=None,
@@ -298,14 +197,10 @@ def _create_web_driver_at_mint_com(
     if session_path is not None:
         chrome_options.add_argument("user-data-dir=%s" % session_path)
 
-    if use_chromedriver_on_path:
-        driver = Chrome(options=chrome_options)
-    else:
-        driver = Chrome(
-            options=chrome_options,
-            executable_path=get_stable_chrome_driver(chromedriver_download_path),
-        )
-    return driver
+    # selenium will default to use the chromedriver in your project directory.
+    # otherwise if the chromedriver binary is not present in project directory,
+    # selenium manager will download one in .cache/selenium
+    return Chrome(options=chrome_options)
 
 
 def sign_in(
@@ -343,10 +238,17 @@ def sign_in(
         expected_conditions.presence_of_element_located(
             (
                 By.CSS_SELECTOR,
-                "#ius-link-use-a-different-id-known-device, #ius-userid, #ius-identifier, #ius-option-username",
+                ".ius-hosted-ui-main-container, "
+                "#ius-link-use-a-different-id-known-device, "
+                "#ius-userid, "
+                "#ius-identifier, "
+                "#ius-option-username, "
+                '[data-testid="IdentifierFirstSubmitButton"], '
+                '[data-testid="AccountChoicesUseDifferentId"]',
             )
         )
     )
+
     driver.implicitly_wait(0)  # seconds
 
     user_selection_page(driver)
@@ -376,6 +278,7 @@ def sign_in(
 
         # Wait until logged in, just in case we need to deal with MFA.
 
+        handle_login_failures(driver)
         if not bypass_verified_user_page(driver):
             # if bypass_verified_user_page was present, then MFA already done
             bypass_passwordless_login_page(driver)
@@ -417,7 +320,10 @@ def sign_in(
 
 
 def home_page(driver):
-    element = driver.find_element(By.LINK_TEXT, "Sign in").click()
+    try:
+        element = driver.find_element(By.LINK_TEXT, "Sign in").click()
+    except WebDriverException:
+        logger.info("WebDriverException when clicking Sign In")
 
 
 def user_selection_page(driver):
@@ -426,7 +332,10 @@ def user_selection_page(driver):
         driver.find_element(By.ID, "ius-link-use-a-different-id-known-device").click()
         WebDriverWait(driver, 20).until(
             expected_conditions.presence_of_element_located(
-                (By.CSS_SELECTOR, "#ius-userid, #ius-identifier, #ius-option-username")
+                (
+                    By.CSS_SELECTOR,
+                    '#ius-userid, #ius-identifier, #ius-option-username, [data-testid="IdentifierFirstSubmitButton"]',
+                )
             )
         )
     except NoSuchElementException:
@@ -470,12 +379,71 @@ def handle_different_page_username_password(driver, email):
                 break
 
 
+def handle_login_failures(driver):
+    try:
+        WebDriverWait(driver, 0).until(
+            expected_conditions.presence_of_element_located(
+                (
+                    By.XPATH,
+                    '//div[contains(text(), "We can\'t find anyone with ")][@id="ius-identifier-first-error"]',
+                )
+            )
+        )
+        raise RuntimeError(
+            "Login to Mint failed: Mint does not recognize your login email"
+        )
+    except TimeoutException:
+        pass
+
+    try:
+        WebDriverWait(driver, 0).until(
+            expected_conditions.presence_of_element_located(
+                (
+                    By.XPATH,
+                    "//div[contains(text(), 'The password you entered is incorrect.')]",
+                )
+            )
+        )
+        raise RuntimeError("Login to Mint failed: incorrect password")
+    except TimeoutException:
+        pass
+
+    try:
+        WebDriverWait(driver, 0).until(
+            expected_conditions.presence_of_element_located(
+                (
+                    By.ID,
+                    "RecaptchaHeader",
+                )
+            )
+        )
+        raise RuntimeError("Login to Mint failed: Captcha presented")
+    except TimeoutException:
+        pass
+
+    try:
+        WebDriverWait(driver, 0).until(
+            expected_conditions.presence_of_element_located(
+                (
+                    By.XPATH,
+                    '//h2[contains(text(), "The feature you\'ve requested is temporarily unavailable")]',
+                )
+            )
+        )
+        raise RuntimeError(
+            "Login to Mint failed: Mint reports that it's temporarily unavailable: you may be blocked."
+        )
+    except TimeoutException:
+        pass
+
+
 def bypass_verified_user_page(driver):
     # bypass "Let's add your current mobile number" interstitial page
     # returns True is page is bypassed
     try:
         skip_for_now = driver.find_element(
-            By.ID, "ius-verified-user-update-btn-skip"
+            By.CSS_SELECTOR,
+            '#ius-verified-user-update-btn-skip, [data-testid="VUUSkipButton"]',
         ).click()
         return True
     except (
@@ -503,7 +471,22 @@ def mfa_selection_page(driver, mfa_method):
 def bypass_passwordless_login_page(driver):
     # bypass "Sign in without a password next time" interstitial page
     try:
-        skip_for_now = driver.find_element(By.ID, "skipWebauthnRegistration").click()
+        skip_for_now = driver.find_element(
+            By.CSS_SELECTOR, "#skipWebauthnRegistration, #signInDifferentWay"
+        ).click()
+    except (
+        NoSuchElementException,
+        StaleElementReferenceException,
+        ElementNotVisibleException,
+        ElementNotInteractableException,
+    ):
+        pass
+    # bypass "Let's make sure you're you" if password login is allowed
+    try:
+        skip_for_now = driver.find_element(
+            By.CSS_SELECTOR,
+            '[data-testid="challengePickerOption_PASSWORD"]',
+        ).click()
     except (
         NoSuchElementException,
         StaleElementReferenceException,
@@ -578,7 +561,9 @@ def search_mfa_method(driver):
                 if result is True:
                     break
         except (NoSuchElementException, ElementNotInteractableException):
-            logger.info("{} MFA Method Not Found".format(constants.MFA_METHOD_LABEL))
+            logger.info(
+                "{} MFA Method Not Found".format(method[constants.MFA_METHOD_LABEL])
+            )
     return mfa_token_input, mfa_token_button, mfa_method
 
 
@@ -636,7 +621,7 @@ def handle_email_by_imap(
         if mfa_code is None:
             mfa_code = (mfa_input_callback or input)(DEFAULT_MFA_INPUT_PROMPT)
         submit_mfa_code(mfa_token_input, mfa_token_button, mfa_code)
-    except (NoSuchElementException, ElementNotInteractableException):
+    except (NoSuchElementException, ElementNotInteractableException, EOFError):
         logger.info("Not on Email MFA Screen")
 
 
@@ -644,7 +629,7 @@ def handle_other_mfa(mfa_token_input, mfa_token_button, mfa_input_callback):
     try:
         mfa_code = (mfa_input_callback or input)(DEFAULT_MFA_INPUT_PROMPT)
         submit_mfa_code(mfa_token_input, mfa_token_button, mfa_code)
-    except (NoSuchElementException, ElementNotInteractableException):
+    except (NoSuchElementException, ElementNotInteractableException, EOFError):
         logger.info("Not on SMS or Authenticator MFA Screen")
 
 
@@ -657,29 +642,42 @@ def submit_mfa_code(mfa_token_input, mfa_token_button, mfa_code):
 def account_selection_page(driver, intuit_account):
     # account selection screen -- if there are multiple accounts, select one
     try:
-        select_account = driver.find_element(By.ID, "ius-mfa-select-account-section")
-        if intuit_account is not None:
-            account_input = select_account.find_element(
-                By.XPATH,
-                "//label/span[text()='{}']/../preceding-sibling::input".format(
-                    intuit_account
-                ),
-            )
-
-            account_input.click()
         WebDriverWait(driver, 20).until(
             expected_conditions.presence_of_element_located(
                 (
                     By.CSS_SELECTOR,
-                    "[data-testid='SelectAccountContinueButton']",
+                    '[data-testid="SelectAccountForm"], [data-testid="IdFirstKnownContainer"]',
                 )
             )
         )
-        mfa_code_submit = driver.find_element(
+        select_account = driver.find_element(
             By.CSS_SELECTOR,
-            '#ius-sign-in-mfa-select-account-continue-btn, [data-testid="SelectAccountContinueButton"]',
+            '[data-testid="SelectAccountForm"], [data-testid="IdFirstKnownContainer"]',
+        )
+        if intuit_account is not None:
+            account_input = select_account.find_element(
+                By.XPATH,
+                "//*/span[text()='{}']/../../../preceding-sibling::input".format(
+                    intuit_account
+                ),
+            )
+            # NOTE: We need to execute a script because simply using account_input.click()
+            #       results in ElementClickInterceptedException.
+            driver.execute_script("arguments[0].click()", account_input)
+
+        WebDriverWait(driver, 20).until(
+            expected_conditions.presence_of_element_located(
+                (
+                    By.CSS_SELECTOR,
+                    "#ius-sign-in-mfa-select-account-continue-btn, [data-testid='SelectAccountContinueButton'], [data-testid='AccountChoiceUsage_0']",
+                )
+            )
+        )
+        driver.find_element(
+            By.CSS_SELECTOR,
+            '#ius-sign-in-mfa-select-account-continue-btn, [data-testid="SelectAccountContinueButton"], [data-testid="AccountChoiceUsage_0"]',
         ).click()
-    except NoSuchElementException:
+    except (TimeoutException, NoSuchElementException):
         logger.info("Not on Account Selection Screen")
 
 
@@ -687,10 +685,12 @@ def password_page(driver, password):
     # password only sometimes after mfa
     try:
         driver.find_element(
-            By.ID, "ius-sign-in-mfa-password-collection-current-password"
+            By.CSS_SELECTOR,
+            "#iux-password-confirmation-password, #ius-sign-in-mfa-password-collection-current-password",
         ).send_keys(password)
         driver.find_element(
-            By.ID, "ius-sign-in-mfa-password-collection-continue-btn"
+            By.CSS_SELECTOR,
+            '#ius-sign-in-mfa-password-collection-continue-btn, [data-testid="passwordVerificationContinueButton"]',
         ).submit()
     except (
         NoSuchElementException,
@@ -710,14 +710,21 @@ def handle_wait_for_sync(driver, wait_for_sync_timeout, fail_if_stale):
                 (By.CSS_SELECTOR, ".AccountStatusBar")
             )
         )
-        WebDriverWait(driver, wait_for_sync_timeout).until(
-            lambda x: "Account refresh complete"
-            in status_web_element.get_attribute("innerHTML")
-        )
+
+        def refresh_complete(x):
+            statusHtml = status_web_element.get_attribute("innerHTML")
+
+            return (
+                ("Account refresh complete" in statusHtml)
+                or ("We can't update your" in statusHtml)
+                or ("need attention" in statusHtml)
+            )
+
+        WebDriverWait(driver, wait_for_sync_timeout).until(refresh_complete)
         return status_web_element.text
     except (TimeoutException, StaleElementReferenceException):
         logger.warning(exceptions.STALE_DATA_ERROR_MESSAGE)
         if fail_if_stale:
             raise exceptions.StaleDataException
-    except (exceptions.StaleDataException):
+    except exceptions.StaleDataException:
         sys.exit(1)
